from __future__ import annotations

import functools
from datetime import datetime

import yaml

from everai.app.app_versioned import AppVersionedMixin
from everai.utils.k8s_sort_key import sort_key
from everai.utils.show_mixin import ShowMixin, TableField
from everai.autoscaling.autoscaler import AutoScaler
from everai.image import Image, BasicAuth
from everai.resource_requests.resource_requests import ResourceRequests
from everai.app.volume_request import VolumeRequest
import typing

from everai.volume import regular_name
from generated.apps import (
    V1App,
    V1AppV1Alpha1,
    V1Metadata,
    V1ImagePullSecrets,
    V1Volume as V1AppVolume,
    VolumeVolumeItem,
    VolumeSecretItem,
    VolumeConfigMapItem,
    V1Resource,
    V1AppSpecV1Alpha1,
    V1AppStatusV1Alpha1, AppSpecV1Alpha1RouteAuthType, V1AppStatusV1Alpha1Status, V1Probe, ProbeHttpGet,
)
from everai.app.app_runner import AppRunnerMixin
from everai.utils.datetime import format_datetime
from generated.apps.models.v1_app_status import V1AppStatus

T = typing.TypeVar('T')


class App(AppRunnerMixin, AppVersionedMixin, ShowMixin):
    _autoscaler: AutoScaler

    table_fields: typing.List[TableField] = [
        TableField('name'),
        TableField('namespace'),
        TableField('status', picker=lambda dt: dt.status.status.value.removeprefix('STATUS_')),
        TableField('workers', picker=lambda dt: f'{dt.status.ready_worker or 0}/{dt.status.desired_worker or 0}'),
        TableField('created_at', formatter=lambda dt: format_datetime(dt)),
    ]

    wide_table_extra_fields: typing.List[TableField] = [
        TableField('updated_at', formatter=lambda dt: format_datetime(dt)),
    ]

    AppVersionedMixin.extend_mapping(
        {
        }
    )

    @staticmethod
    def _get_image_pull_secret(image: typing.Optional[Image]) -> typing.Optional[V1ImagePullSecrets]:
        if image is None or image.auth is None:
            return None
        if not isinstance(image.auth, BasicAuth):
            raise ValueError(f'unsupported auth method {type(image.auth)}')

        return V1ImagePullSecrets(
            username=image.auth.username.to_proto(),
            password=image.auth.password.to_proto(),
        )

    @staticmethod
    def _get_volumes(
            volume_requests: typing.Optional[typing.List[VolumeRequest]] = None,
            secret_requests: typing.Optional[typing.List[str]] = None,
            configmap_requests: typing.Optional[typing.List[str]] = None,
    ) -> typing.List[V1AppVolume]:
        volumes_volumes = [
            V1AppVolume(
                name=regular_name(v.name),
                volume=VolumeVolumeItem(volume=v.name),
            )
            for v in volume_requests or []
        ]
        secrets_volumes = [
            V1AppVolume(
                name=v,
                secret=VolumeSecretItem(secret_name=v),
            )
            for v in secret_requests or []
        ]
        configmap_requests = [
            V1AppVolume(
                name=v,
                config_map=VolumeConfigMapItem(name=v),
            )
            for v in configmap_requests or []
        ]
        result = sum([volumes_volumes, secrets_volumes, configmap_requests], [])
        return None if result is None or len(result) == 0 else result

    def __init__(self,
                 name: str,
                 resource_requests: ResourceRequests,
                 namespace: str = 'default',
                 labels: typing.Optional[typing.Dict[str, str]] = None,
                 image: typing.Optional[Image] = None,
                 autoscaler: typing.Optional[AutoScaler] = None,
                 secret_requests: typing.Optional[typing.List[str]] = None,
                 configmap_requests: typing.Optional[typing.List[str]] = None,
                 volume_requests: typing.Optional[typing.List[VolumeRequest]] = None,
                 *args, **kwargs):
        self._autoscaler = autoscaler

        image_pull_secrets = App._get_image_pull_secret(image)
        volumes = App._get_volumes(volume_requests, secret_requests, configmap_requests)

        self._app = V1App(
            app_v1alpha1=V1AppV1Alpha1(
                metadata=V1Metadata(
                    name=name,
                    namespace=namespace,
                    labels=labels,
                ),
                spec=V1AppSpecV1Alpha1(
                    image=image.image if image is not None else None,
                    image_pull_secrets=image_pull_secrets,
                    volume_mounts=None,
                    env=None,
                    command=None,
                    port=None,
                    readiness_probe=V1Probe(http_get=ProbeHttpGet(path='/-everai-/healthy')),
                    volumes=volumes,
                    resources=resource_requests.to_proto() if resource_requests is not None else None,
                    route_auth_type=None,
                    route_public_key=None,
                ),
                status=None,
            )
        )

    @staticmethod
    def from_proto(v1app: V1App) -> App:
        app = App(name='', resource_requests=ResourceRequests())
        app._app = v1app
        if app.metadata.namespace is None or app.metadata.namespace == '':
            app.metadata.namespace = 'default'
        return app

    def to_proto(self) -> V1App:
        return self._app

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        data = self.get_to_dict()
        data.update(dict(
            version="everai/v1alpha1",
            kind="App",
        ))

        return dict(sorted(data.items(), key=lambda item: sort_key(item[0])))

    @staticmethod
    def from_dict(data: typing.Dict[str, typing.Any]) -> App:
        version = data.pop('version', None)
        if version is None:
            raise ValueError("version is required")

        kind = data.pop("kind", None)
        if kind is None:
            raise ValueError("version is required")
        if kind != 'App':
            raise ValueError('kind should be "App"')

        if version == "everai/v1alpha1":
            data = App.convert_from_dict(data, "v1alpha1")

            v1alpha1 = V1AppV1Alpha1.from_dict(data)
            return App.from_proto(V1App(app_v1alpha1=v1alpha1))
        else:
            raise ValueError(f"unsupported version {version}")

    def to_yaml(self) -> str:
        d = self.to_dict()
        return yaml.dump(d, default_flow_style=False, sort_keys=False)

    @staticmethod
    def from_yaml_file(filename: str) -> App:
        with open(filename, 'r') as f:
            return App.from_yaml_stream(f)

    @staticmethod
    def from_yaml_stream(stream: typing.IO) -> App:
        data = yaml.safe_load(stream)
        return App.from_dict(data)

    @property
    def volume_requests(self) -> typing.List[VolumeRequest]:
        return [VolumeRequest(x.volume.volume, False) for x in self.spec.volumes if x.volume is not None]

    @property
    def secret_requests(self) -> typing.List[str]:
        return [x.secret.secret_name for x in self.spec.volumes if x.secret is not None]

    @property
    def configmap_requests(self) -> typing.List[str]:
        return [x.config_map.name for x in self.spec.volumes if x.config_map is not None]
